@@ -1,5 +1,6 @@
 /*****************************************************************************
  * Authors: Louis Bavoil <bavoil@cs.utah.edu>
+ *          Peter Leese <peter@leese.net>
  *          Fredrik Hbinette <hubbe@hubbe.net>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -19,9 +20,11 @@
 
 #define _GNU_SOURCE /* for getsid() */
 #include "mozplugger.h"
+#ifndef __sun
 #include <stdint.h>
+#endif
 
-static Display *display;
+static Display * display = 0;
 static int pipe_fd;
 static int flags;
 static int repeats;
@@ -81,7 +84,7 @@ uint32_t getHostId()
        effecting the swallow of victims is so infinitimisally small! */
 
     id = 0;
-    for(i=0; i < sizeof(hostName)/sizeof(uint32_t); i++)
+    for(i=0; i < (int)(sizeof(hostName)/sizeof(uint32_t)); i++)
     {
         id = ((id << 5) ^ (id >> 27)) ^ ((uint32_t *)hostName)[i];
     }
@@ -110,7 +113,8 @@ static int getSwallowMutexOwner(uint32_t * hostId, uint32_t * pid)
 	if((type != XA_INTEGER) || (fmt!=32) || (nitems!=2) || (bytes != 0))
         {
 	    fprintf(stderr, "XGetWindowProperty returned bad values " 
-                            "%ld,%d,%ld,%ld\n", type, fmt, nitems, bytes);
+                            "%ld,%d,%lu,%lu\n", (long) type, fmt, nitems, 
+                            bytes);
         }
 	else
 	{
@@ -125,11 +129,12 @@ static int getSwallowMutexOwner(uint32_t * hostId, uint32_t * pid)
 
 static void setSwallowMutexOwner(uint32_t hostId, uint32_t pid)
 {
-    D("Setting swallow mutex owner, hostId = 0x%08X, pid=%d\n", hostId, pid); 
     unsigned long temp[2] = {hostId, pid};
+    D("Setting swallow mutex owner, hostId = 0x%08X, pid=%d\n", hostId, pid); 
     XChangeProperty(display, wattr.root, swallowMutex,
                         XA_INTEGER, 32, PropModeReplace,
                         (unsigned char*) (&temp), 2);
+    XSync(display, False);
 }
  
 static void takeSwallowMutex()
@@ -172,7 +177,7 @@ static void takeSwallowMutex()
                    if process exists. */
             	if( (getsid(otherPid) < 0 ) && (errno == ESRCH))
 	        {
-	    	    fprintf(stderr,"Strange other Pid(%d) cannot be found\n", 
+	    	    fprintf(stderr,"Strange other Pid(%lu) cannot be found\n", 
                                    otherPid);
                     break;    /* OK force early exit of inner while loop */
                 }
@@ -243,8 +248,8 @@ static void adjust_window_size(void)
 {
      int x = 0;
      int y = 0;
-     int w = windata.width;
-     int h = windata.height;
+     int w = (int) windata.width;
+     int h = (int) windata.height;
 
      int tmpw, tmph;
 
@@ -282,7 +287,7 @@ static void adjust_window_size(void)
   
      D("New size: %dx%d+%d+%d\n", w, h, x, y);
      
-     XMoveResizeWindow(display, victim, x, y, w, h);
+     XMoveResizeWindow(display, victim, x, y, (unsigned)w, (unsigned)h);
 }
 
 /*****************************************************************************
@@ -337,7 +342,10 @@ static int is_viewable(Display *display, Window w)
      }
      return 1;
 }
-
+/******************************************************************************
+ * Traditionally strcmp returns -1, 0 and +1 depending if name comes before
+ * or after windowname, this function also returns +1 if error
+ *****************************************************************************/
 static int my_strcmp(char *windowname, char *name)
 {
      if (!name) return 1;
@@ -352,6 +360,8 @@ static int my_strcmp(char *windowname, char *name)
      case '*':
 	  return strncasecmp(name+1, windowname, strlen(name)-1);
      default:
+          /* Return 0 for success so need to invert logic as strstr
+             returns pointer to the match or NULL if no match */
 	  return !strstr(windowname, name);
      }
 }
@@ -367,7 +377,7 @@ static char check_window_name(Window w, char *name, int (*cb)(Window))
      {
 	  D("Winrecur, checking window NAME %x (%s != %s)\n",
 	    w, windowname, name);
-	  match = !my_strcmp(windowname, name);
+	  match = (my_strcmp(windowname, name) == 0);
 	  XFree(windowname);
 	  if (match) {
 	       return cb(w);
@@ -379,7 +389,7 @@ static char check_window_name(Window w, char *name, int (*cb)(Window))
      {
 	  D("Winrecur, checking window CLASS %x (%s != %s)\n",
 	    w, windowclass.res_name, name);
-	  match = !my_strcmp(windowclass.res_name, name);
+	  match = (my_strcmp(windowclass.res_name, name) == 0);
 	  XFree(windowclass.res_class);
 	  XFree(windowclass.res_name);
 	  if (match) {
@@ -393,7 +403,7 @@ static Window winrecur(Window from, int depth, char *name, int (*cb)(Window))
 {
      Window root, parent;
      Window *children;
-     int e;
+     unsigned int e;
      unsigned int num_children;
 
      D("Winrecur, checking window %x (depth=%d)\n",from,depth);
@@ -480,25 +490,71 @@ static int setup_display()
      return 0;
 }
 
+/*****************************************************************************
+ * Wrapper for execlp() that calls the application.
+ *
+ * WARNING: This function runs in the daughter process so one must assume the
+ * daughter uses a copy (including) heap memory of the parent's memory space
+ * i.e. any write to memory here does not affect the parent memory space. 
+ * Since Linux uses copy-on-write, best leave memory read-only and once execlp
+ * is called, all daughter memory is wiped anyway (except the stack).
+ *****************************************************************************/
 static void run_app(char **argv)
 {
-     int nl;
-
-     if (flags & H_NOISY)
+     /* Close all File descriptors inherited from the parent,
+      * that's the XServer connection FD and pipe_fd */
+     if(display)
      {
-	  D("Redirecting stdout and stderr\n");
-	  if ((nl = open("/dev/null", O_RDONLY)) == -1)
-	       exit(EX_UNAVAILABLE);
-	  dup2(nl, 1);
-	  dup2(nl, 2);
-	  close(nl);
+          close(ConnectionNumber(display));
      }
-     
+     close(pipe_fd);
+
+     /* Redirect stdout & stderr to /dev/null */
+     if(flags & (H_NOISY | H_DAEMON))
+     {
+	  const int ofd = open("/dev/null", O_RDONLY);
+
+	  D("Redirecting stdout and stderr\n");
+
+	  if(ofd == -1)
+          {
+	       exit(EX_UNAVAILABLE);
+          }
+
+	  dup2(ofd, 1);
+	  dup2(ofd, 2);
+	  close(ofd);
+     }
+
+     /* For a DAEMON we must also redirect STDIN to /dev/null
+      * otherwise we can get some weird behaviour especially with realplayer
+      * (I suspect it uses stdin to communicate with its peers) */
+     if(flags & H_DAEMON)
+     {
+          const int ifd = open("/dev/null", O_WRONLY);
+
+          D("Redirecting stdin also\n");
+
+          if(ifd == -1)
+          {
+               exit(EX_UNAVAILABLE);
+          }
+          dup2(ifd, 0);
+          close(ifd);
+     }
+#if DEBUG
+     close_debug();
+#endif
      execvp(argv[0], argv);
      D("Execvp failed. (errno=%d)\n", errno);
      exit(EX_UNAVAILABLE);
 }
 
+
+/*****************************************************************************
+ * Looks for the window and reparents it when found.
+ *
+ *****************************************************************************/
 static void find_victim(Window window)
 {
      if (!victim)
@@ -690,28 +746,32 @@ static void check_all_events()
      }
 }
 
-static void handle_app(int pid)
+static void handle_app(pid_t pid)
 {
      int status;
 
      if (flags & H_SWALLOW)
      {
+          /* Whilst waiting for the Application to complete, check X events */
 	  while (!waitpid(pid, &status, WNOHANG))
 	  {
 	       check_all_events();
 	  }
-          giveSwallowMutex();      // Make sure the semaphore has been released
+          /* Make sure the semaphore has been released */
+          giveSwallowMutex();
+     }
+     else if(flags & H_DAEMON)
+     {
+          /* If Daemon, then it is not supposed to exit, so we exit instead */
+          exit(0);
      }
      else
      {
+          /* Just wait for the Application to complete dont check X events */
 	  waitpid(pid, &status, 0);
      }
 
-     if (flags & H_DAEMON)
-     {
-	  exit(0);
-     }
-
+     /* If Application completed is a bad way, then lets give up now */
      if (!WIFEXITED(status))
      {
 	  D("Process dumped core or something...\n");
@@ -727,6 +787,10 @@ static void handle_app(int pid)
      D("Exited OK!\n");
 }
 
+/******************************************************************************
+ * main() normally called from the child process started by mozplugger.so 
+ * 
+ *****************************************************************************/
 int main(int argc, char **argv)
 {
      unsigned long temp = 0;
@@ -735,7 +799,6 @@ int main(int argc, char **argv)
      if (argc < 2)
      {
 	  fprintf(stderr,"MozPlugger version " VERSION " helper application.\n"
-		  "Written by Fredrik Hubinette and Louis Bavoil.\n" 
 		  "Please see 'man mozplugger' for details.\n");
 	  exit(1);
      }
@@ -768,7 +831,7 @@ int main(int argc, char **argv)
      {
 	  char *argv[10];
 	  int loops = 1;
-	  int pid;
+	  pid_t pid;
 
 	  /* This application will use the $repeat variable */
 	  if (flags & H_REPEATCOUNT)
@@ -779,41 +842,30 @@ int main(int argc, char **argv)
 	       loops = MAXINT;
 
 	  /* Create command line */
-	  if (flags & H_CONTROLS)
-	  {
-	       argv[0] = getenv("controller");
-	       argv[1] = command;
-	       argv[2] = NULL;
-	  }
-	  else
-	  {
-	       argv[0] = "/bin/sh";
-	       argv[1] = "-c";
-	       argv[2] = command;
-	       argv[3] = NULL;
-	  }
+	  argv[0] = "/bin/sh";
+	  argv[1] = "-c";
+	  argv[2] = command;
+	  argv[3] = NULL;
 	  
 	  if (setup_display())
 	  {
 	       D("setup_display() failed\n");
 	       flags &=~ H_SWALLOW;
 	  }
-	  else
+
+	  if (flags & H_SWALLOW)
 	  {
-              if(flags & H_SWALLOW)
-              {
-                  /* If we are swallowing a victim window we need to guard
-                     against more than one instance of helper running in
-                     parallel, this is done using a global mutex semaphore. */
-                  initSwallowMutex();
+	       /* If we are swallowing a victim window we need to guard
+		  against more than one instance of helper running in
+		  parallel, this is done using a global mutex semaphore. */
+	       initSwallowMutex();
 
-                  /* Belt & braces, make sure we give semaphore on error. */
-                  signal(SIGTERM, giveSwallowMutex);
+	       /* Belt & braces, make sure we give semaphore on error. */
+	       signal(SIGTERM, giveSwallowMutex);
 
-                  takeSwallowMutex(); 
-               }
+	       takeSwallowMutex(); 
 
-               /* Has to be done before forking */
+	       /* Has to be done before forking */
 	       save_initial_windows();
 	       XSelectInput(display, WINDOW, SubstructureRedirectMask);
 	       XSelectInput(display, wattr.root, SubstructureNotifyMask);
@@ -825,11 +877,16 @@ int main(int argc, char **argv)
                giveSwallowMutex();  
 	       exit(EX_UNAVAILABLE);
           }
-      
-	  if (!pid)
+     
+	  if (pid == 0)
 	  {
-	       XSelectInput(display, WINDOW, 0);
-	       XSelectInput(display, wattr.root, 0);
+               /* Child process */
+
+#if 0 /* Peter Leese - I'm sure this is not needed, especially if FD is closed 
+         in run_app() */
+               XSelectInput(display, WINDOW, 0);
+               XSelectInput(display, wattr.root, 0);
+#endif
 	       D("Running %s\n", command);
 	       run_app(argv);
 	  }
@@ -841,6 +898,10 @@ int main(int argc, char **argv)
 	       if (repeats < MAXINT)
 		    repeats -= loops;
 	  }
+          /* On each repeat we re-open the XServer connection, so need to
+           * close it at end of loop - TODO perhaps we should keep in open
+           * for all repeats? */
+          XCloseDisplay(display);
      }
 
      exit(0);
