@@ -1,5 +1,6 @@
 /*****************************************************************************
  * Authors: Louis Bavoil <bavoil@cs.utah.edu>
+ *          Peter Leese <peter@leese.net>
  *          Fredrik Hbinette <hubbe@hubbe.net>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -39,13 +40,13 @@ typedef struct data
      Display *display;
      char *displayname;
      NPWindow windata;
-     int pid;
+     pid_t pid;
      int fd;
      int repeats;
-     int flags;
+     int cmd_flags;
      char *command;
      char *winname;
-     uint16 mode;
+     int mode_flags;
      char *mimetype;
      char *href;
      char *mms;
@@ -89,21 +90,47 @@ static char *helper_fname;
 static char *controller_fname;
 static int num_handlers;
 static handler_t handlers[MAX_NUM_HANDLERS];
+static int browserApiMajorVer;
+static int browserApiMinorVer;
 
 /*****************************************************************************
  * Wrapper for fork() which handles some magic needed to prevent netscape
- * from freaking out. It also prevents interferrance from signals.
+ * from freaking out. It also prevents interference from signals.
+ * Also it closes all those file descriptors inherited from the parent 
+ * (except STDIN, STDOUT & STDERR (0,1,2))
+ *
  *****************************************************************************/
-static int my_fork(NPP instance)
+static pid_t my_fork(NPP instance, int pipeFd)
 {
-     int pid;
-     sigset_t set,oset;
-     int signum;
+     pid_t pid;
+     sigset_t set;
+     sigset_t oset;
+     int maxFds;
 
      /* Mask all the signals to avoid being interrupted by a signal */
      sigfillset(&set);
      sigprocmask(SIG_SETMASK, &set, &oset);
 
+#if 0      /* If sysconf not support do this? */
+     {
+          struct rlimit rl;
+          if (getrlimit(RLIMIT_NOFILE, &fdl) < 0)
+          {
+               maxFds = 256;
+          }
+          else if (rl.rlim_max == RLIM_INFINITY)
+          {
+               maxFds = 1024;
+          }
+          else
+          {
+              maxFds = rl.rlim_max;
+          }
+     }
+#else
+     maxFds = sysconf(_SC_OPEN_MAX);
+#endif
+
      D(">>>>>>>>Forking<<<<<<<<,\n");
 
      if ((pid = fork()) == -1)
@@ -111,20 +138,35 @@ static int my_fork(NPP instance)
 	  return -1;
      }
 
-     if (!pid)
+     if (pid == 0)
      {
+          int i;
+          int signum;
+
 	  alarm(0);
 
-	  if (!(THIS->flags & H_DAEMON))
+	  if (!(THIS->cmd_flags & H_DAEMON))
 	       setsid();
 
 	  for (signum=0; signum < NSIG; signum++)
 	       signal(signum, SIG_DFL);
 
-	  THIS->display = NULL;
-
+          /* Close all those File descriptors inherited from the
+             parent (except STDIN, STDOUT, STDERR and pipeFd. */
 #ifdef DEBUG
 	  close_debug();
+#endif
+          D("Closing up to %i Fds\n", maxFds);
+          for (i = 3; i < pipeFd; i++)
+          {
+               close(i);
+          }
+          for (i = pipeFd+1; i < maxFds; i++)
+          {
+               close(i);
+          }
+
+#ifdef DEBUG
 	  D("Child re-opened ndebug\n");
 #endif
      }
@@ -142,41 +184,51 @@ static int my_fork(NPP instance)
 /*****************************************************************************
  * Wrapper for putenv().
  *****************************************************************************/
-static void my_putenv(char *buffer, int *offset, const char *var, const char *value)
+static void my_putenv(char *buffer, int *offset, const char *var, 
+                                                 const char *value)
 {
      int l = strlen(var) + strlen(value) + 2;
-     if (!(*offset + l < ENV_BUFFER_SIZE))
+     if (*offset + l >= ENV_BUFFER_SIZE)
      {
 	  D("Buffer overflow in putenv(%s=%s)\n", var, value);
 	  return;
      }
      
-     sprintf(buffer+*offset, "%s=%s", var, value);
+     snprintf(buffer+*offset, l, "%s=%s", var, value);
      putenv(buffer+*offset);
      *offset += l;
 }
 
 /*****************************************************************************
  * Wrapper for execlp() that calls the helper.
+ *
+ * WARNING: This function runs in the daughter process so one must assume the
+ * daughter uses a copy (including) heap memory of the parent's memory space
+ * i.e. any write to memory here does not affect the parent memory space. 
+ * Since Linux uses copy-on-write, best leave memory read-only and once execlp
+ * is called, all daughter memory is wiped anyway (except the stack).
  *****************************************************************************/
-static void run(NPP instance, const char *file)
+static void run(NPP instance, const char *file, int fd)
 {
      char buffer[ENV_BUFFER_SIZE];
      char foo[SMALL_BUFFER_SIZE];
      int offset = 0;
      int i;
+     int flags = THIS->cmd_flags;
+     int autostart = THIS->autostart;
 
-     if ((THIS->flags & H_CONTROLS) && (THIS->mode == NP_FULL))
-	  THIS->flags &= ~(H_CONTROLS | H_SWALLOW | H_FILL);
+     /* Specially linking objects to external app must use controls */
+     if (THIS->mode_flags & H_LINKS)
+	  flags |= H_CONTROLS;
 
-     // If No autostart seen and using controls, dont autostart by default
-     if ((THIS->flags & H_CONTROLS) && (THIS->autostartNotSeen))
-	  THIS->autostart = 0;
+     /* If no autostart seen and using controls, dont autostart by default */
+     if ((flags & H_CONTROLS) && (THIS->autostartNotSeen))
+	  autostart = 0;
 
      snprintf(buffer, sizeof(buffer), "%d,%d,%d,%lu,%d,%d,%d,%d",
-	      THIS->flags,
+	      flags,
 	      THIS->repeats,
-	      THIS->fd,
+	      fd,
 	      (unsigned long int) THIS->windata.window,
 	      (int) THIS->windata.x,
 	      (int) THIS->windata.y,
@@ -212,7 +264,7 @@ static void run(NPP instance, const char *file)
      
      my_putenv(buffer, &offset, "file", file);
 
-     my_putenv(buffer, &offset, "autostart", THIS->autostart ? "1" : "0");
+     my_putenv(buffer, &offset, "autostart", autostart ? "1" : "0");
       
      if (THIS->winname)
 	  my_putenv(buffer, &offset, "winname", THIS->winname);
@@ -220,15 +272,18 @@ static void run(NPP instance, const char *file)
      if (THIS->displayname)
 	  my_putenv(buffer, &offset, "DISPLAY", THIS->displayname);
 
-     if (controller_fname)
-	  my_putenv(buffer, &offset, "controller", controller_fname);
-
      for (i = 0; i < THIS->num_arguments; i++)
 	  if (THIS->args[i].value)
 	       my_putenv(buffer, &offset, THIS->args[i].name, THIS->args[i].value);
 
-     execlp(helper_fname, helper_fname, buffer, THIS->command, NULL);
-
+     if(flags & H_CONTROLS)
+     {
+          execlp(controller_fname, helper_fname, buffer, THIS->command, NULL);
+     }
+     else
+     {
+          execlp(helper_fname, helper_fname, buffer, THIS->command, NULL);
+     }
      D("EXECLP FAILED!\n");
      
      _exit(EX_UNAVAILABLE);
@@ -372,7 +427,9 @@ static int parse_flags(char **x, command_t *commandp)
 	  { "noisy",		H_NOISY 	},
 	  { "embed",            H_EMBED         },
 	  { "noembed",          H_NOEMBED       },
+          { "links",            H_LINKS         },
 	  { "hidden",           H_HIDDEN        },
+          { "controls",         H_CONTROLS      },
 	  { NULL, 		0 		}
      };
      flag_t *f;
@@ -391,13 +448,6 @@ static int parse_flags(char **x, command_t *commandp)
 	       if ((*x = get_winname(*x, "swallow", commandp)))
 		    return 1;
 	  }
-	  if (match_word(*x, "controls"))
-	  {
-	       commandp->flags |= H_CONTROLS | H_FILL | H_SWALLOW | H_HIDDEN;
-	       strcpy(commandp->winname, "mozplugger-controller");
-	       *x += 8;
-	       return 1;
-	  }
      }
      return 0;
 }
@@ -597,7 +647,7 @@ static int find_helper_file(char *basename, int (*cb)(char *,void *data), void *
 static int read_config_cb(char *fname, void *data)
 {
      int m4out[2];
-     int pid;
+     pid_t pid;
      int fd;
      FILE *fp;
 
@@ -729,20 +779,22 @@ static int safeURL(char* url)
 __inline
 static int match_command(NPP instance, int streaming, command_t *c)
 {
-     char embedded = (THIS->mode == NP_EMBED);
+#define MODE_MASK (H_NOEMBED | H_EMBED | H_LINKS)
 
      D("Checking command: %s\n", c->cmd);
 
-     if (embedded && (c->flags & H_NOEMBED))
+     /* If command is specific to a particular mode... */
+     if (c->flags & MODE_MASK)
      {
-	  D("Flag mismatch: embed\n");
-	  return 0;
-     }
-     if ((!embedded) && (c->flags & H_EMBED))
-     {
-	  D("Flag mismatch: noembed\n");
-	  return 0;
+          /* Check it matches the current mode... */
+          if ( (THIS->mode_flags & MODE_MASK) != (c->flags & MODE_MASK) )
+          {
+	       D("Flag mismatch: mode different %x != %x\n",
+                     THIS->mode_flags & MODE_MASK,  c->flags & MODE_MASK);
+	       return 0;
+          }
      }
+
      if ((c->flags & H_LOOP) && (THIS->repeats != MAXINT))
      {
 	  D("Flag mismatch: loop\n");
@@ -766,7 +818,8 @@ static int match_mime_type(NPP instance, mimetype_t *m)
      sscanf(mimetype, "%s", mimetype);
 
      D("Checking '%s' ?= '%s'\n", mimetype, THIS->mimetype);
-     if (strcasecmp(mimetype, THIS->mimetype) && strcmp(mimetype,"*"))
+     if ((strcasecmp(mimetype, THIS->mimetype) != 0) 
+             && (strcmp(mimetype,"*") != 0))
      {
 	  D("Not same.\n");
 	  return 0;
@@ -796,7 +849,7 @@ static int match_handler(handler_t *h, NPP instance, int streaming)
 		    c = &h->cmds[cid];
 		    if (match_command(instance, streaming, c))
 		    {
-			 THIS->flags = c->flags;
+			 THIS->cmd_flags = c->flags;
 			 THIS->command = c->cmd;
 			 THIS->winname = c->winname;
 			 return 1;
@@ -909,11 +962,7 @@ NPError NPP_GetValue(void *instance, NPPVariable variable, void *value)
 	  snprintf(desc_buffer, sizeof(desc_buffer),
 		   "MozPlugger version "
 		   VERSION
-		   ", written by "
-		   "<a href=http://fredrik.hubbe.net/>Fredrik H&uuml;binette</a> "
-		   "<a href=mailto:hubbe@hubbe.net>&lt;hubbe@hubbe.net&gt</a> "
-		   "and Louis Bavoil "
-		   "<a href=mailto:bavoil@cs.utah.edu>&lt;bavoil@cs.utah.edu&gt</a>.<br>"
+		   ", written by Fredrik H&uuml;binette, Louis Bavoil and Peter Leese.<br>"
 		   "For documentation on how to configure mozplugger, "
 		   "check the man page. (type <tt>man&nbsp;mozplugger</tt>)"
 		   " <table> "
@@ -969,24 +1018,34 @@ NPError NPP_New(NPMIMEType pluginType,
 		NPSavedData* saved)
 {
      int e;
-     int src_idx = -1;
 
-     D("NEW (%s)\n", pluginType);
+     int src_idx = -1;
+     int href_idx = -1;
+     int data_idx = -1;
+     int alt_idx = -1;
+     int autostart_idx = -1;
+     int autohref_idx = -1;
+     int target_idx = -1;
+
+     char *url = NULL;
+
+     D("NEW (%s) - instance=%p\n", pluginType, instance);
 
      if (!instance)
      {
-	  D("Invalid instance pointer\n");
 	  return NPERR_INVALID_INSTANCE_ERROR;
      }
 
      if (!pluginType)
      {
-	  D("Invalid plugin type\n");
 	  return NPERR_INVALID_INSTANCE_ERROR;
      }
 
      THIS = NPN_MemAlloc(sizeof(data_t));
-     if (!THIS) return NPERR_OUT_OF_MEMORY_ERROR;
+     if (!THIS) 
+          return NPERR_OUT_OF_MEMORY_ERROR;
+
+
      memset((void *)THIS, 0, sizeof(data_t));
 
      THIS->windata.window = 0;
@@ -996,61 +1055,165 @@ NPError NPP_New(NPMIMEType pluginType,
      THIS->repeats = 1;
      THIS->autostart = 1;
      THIS->autostartNotSeen = 1;
-     THIS->mode = mode;
+     if(mode == NP_EMBED)
+     {
+        THIS->mode_flags = H_EMBED;
+     }
+     else
+     {
+        THIS->mode_flags = H_NOEMBED;
+     }
 
      if (!(THIS->mimetype = strdup(pluginType)))
 	  return NPERR_OUT_OF_MEMORY_ERROR;
 
      THIS->num_arguments = argc;
-     if (!(THIS->args = (argument_t *)NPN_MemAlloc(sizeof(argument_t) * argc)))
+     if (!(THIS->args = (argument_t *)NPN_MemAlloc(
+                                          (uint32)(sizeof(argument_t) * argc))))
 	  return NPERR_OUT_OF_MEMORY_ERROR;
      
      for (e = 0; e < argc; e++)
      {
-	  if (!strcasecmp("loop", argn[e]))
+	  if (strcasecmp("loop", argn[e]) == 0)
 	  {
 	       THIS->repeats = my_atoi(argv[e], MAXINT, 1);
 	  }
-	  else if (!strcasecmp("autostart", argn[e]))
+          /* realplayer also uses numloop tag */
+          /* windows media player uses playcount */
+          else if((strcasecmp("numloop", argn[e]) == 0) ||
+                  (strcasecmp("playcount", argn[e]) == 0))
+          {
+	       THIS->repeats = atoi(argv[e]);
+          }
+	  else if((strcasecmp("autostart", argn[e]) == 0) ||
+	          (strcasecmp("autoplay", argn[e]) == 0))
 	  {
-	       THIS->autostart = !!my_atoi(argv[e], 1, 0);
-	       THIS->autostartNotSeen = 0;
+               autostart_idx = e;
 	  }
-	  /* get the index of the src attribute */
-	  else if (!strcasecmp("src", argn[e]))
+	  /* get the index of the src attribute if this is a 'embed' tag */
+	  else if (strcasecmp("src", argn[e]) == 0)
 	  {
 	       src_idx = e;
 	  }
-	  /* copy the flag to put it into the environement later */
-	  D("VAR_%s=%s\n", argn[e], argv[e]);
-	  if (!(THIS->args[e].name = (char *)malloc(strlen(argn[e]) + 5)))
-	       return NPERR_OUT_OF_MEMORY_ERROR;
-	  sprintf(THIS->args[e].name, "VAR_%s", argn[e]);
-	  THIS->args[e].value = argv[e] ? strdup(argv[e]) : NULL;
-     }
-
-     /* Special case for quicktime. If there's an href or qtsrc attribute, we 
-	want that instead of src but we HAVE to have a src first. */
-     for (e = 0; e < argc; e++)
-     {
-          D("arg %d(%s): %s\n",e,argn[e],argv[e]);
-          if ((!strcasecmp("href", argn[e]) ||
-	       !strcasecmp("qtsrc", argn[e])) &&
-	      src_idx > -1 && !THIS->href)
+	  /* get the index of the data attribute if this is a 'object' tag */
+          else if (strcasecmp("data", argn[e]) == 0)
           {
-	       if (!(THIS->href = strdup(argv[e])))
-		    return NPERR_OUT_OF_MEMORY_ERROR;
+               data_idx = e;
+          }
+          /* Special case for quicktime. If there's an href or qtsrc attribute,
+           * remember it for now */
+          else if (((strcasecmp("href", argn[e]) == 0) ||
+	            (strcasecmp("qtsrc", argn[e]) == 0)) &&
+	            href_idx == -1)
+          {
+               href_idx = e;
+          }
+          else if (((strcasecmp("filename", argn[e]) == 0) ||
+	            (strcasecmp("url", argn[e]) == 0) ||
+	            (strcasecmp("location", argn[e]) == 0)) &&
+                    alt_idx == -1)
+          {
+               alt_idx = e;
+          }
+          /* Special case for quicktime. If there's an autohref or target
+           * attributes remember them for now */
+          else if (strcasecmp("target", argn[e]) == 0)
+          {
+               target_idx = e;
+          }
+	  else if(strcasecmp("autohref", argn[e]) == 0)
+	  {
+               autohref_idx = e;
+	  }
+
+	  /* copy the flag to put it into the environment later */
+	  D("VAR_%s=%s\n", argn[e], argv[e]);
+          {
+               const int len = strlen(argn[e]) + 5;
+
+    	       if (!(THIS->args[e].name = (char *)malloc(len)))
+	             return NPERR_OUT_OF_MEMORY_ERROR;
+	       snprintf(THIS->args[e].name, len, "VAR_%s", argn[e]);
+ 	       THIS->args[e].value = argv[e] ? strdup(argv[e]) : NULL;
           }
      }
 
-     if (src_idx > -1)
+     if (src_idx >= 0)
      {
-	  char *url = argv[src_idx];
-	  if (!strncmp(url, "mms://", 6) || !strncmp(url, "mmst://", 7))
+          url = THIS->args[src_idx].value;
+          /* Special case for quicktime. If there's an href or qtsrc 
+           * attribute, we want that instead of src but we HAVE to
+           * have a src first. */
+          if (href_idx >= 0)
+          {
+	       D("Special case QT detected\n");
+	       THIS->href = THIS->args[href_idx].value;
+
+               autostart_idx = autohref_idx;
+
+               if(target_idx >= 0)
+               {
+                   /* One of those clickable Quicktime linking objects! */
+                   THIS->mode_flags &= ~(H_EMBED | H_NOEMBED);
+                   THIS->mode_flags |= H_LINKS;
+               }
+
+          }
+     }
+     else if (data_idx >= 0)
+     {
+          D("Looks like an object tag with data attribute\n");
+          url = THIS->args[data_idx].value;
+     }
+     else if (alt_idx >= 0)
+     {
+          D("Fall-back use alternative tags\n");
+          url = THIS->args[alt_idx].value;
+     }
+
+     /* Do the autostart check here, AFTER we have processed the QT special
+      * case which can change the autostart attribute */
+     if(autostart_idx > 0)
+     {
+	  THIS->autostart = !!my_atoi(argv[autostart_idx], 1, 0);
+	  THIS->autostartNotSeen = 0;
+     }
+
+     if (url)
+     {
+          /* Mozilla does not support the following protocols directly and
+           * so it never calls NPP_NewStream for these protocols */
+	  if(   (strncmp(url, "mms://", 6) == 0)        
+             || (strncmp(url, "mmsu://", 7) == 0)    /* MMS over UDP */
+             || (strncmp(url, "mmst://", 7) == 0)    /* MMS over TCP */
+             || (strncmp(url, "rtsp://", 7) == 0)
+             || (strncmp(url, "rtspu://", 8) == 0)   /* RTSP over UDP */
+             || (strncmp(url, "rtspt://", 8) == 0))  /* RTSP over TCP */
 	  {
-	       D("Detected MMS\n");
-	       THIS->mms = strdup(url);
+	       D("Detected MMS -> url=%s\n", url);
+	       THIS->mms = url;
 	  }
+          else 
+          {
+               /* For protocols that Mozilla does support, sometimes
+                * the browser will call NPP_NewStream straight away, some
+                * times it wont (depends on the nature of the tag). So that
+                * it works in all cases call NPP_GetURL, this may result
+                * in NPP_NewStream() being called twice (i.e. if this is an
+                * embed tag with src attribute or object tag with data
+                * attribute) */
+               if (mode == NP_EMBED)
+               {
+                    const NPError retVal = NPN_GetURL(instance, url, 0);
+                    if(retVal != NPERR_NO_ERROR)
+                    {
+                         D("NPN_GetURL(%s) failed with %i\n", url, retVal);
+
+	                 fprintf(stderr, "MozPlugger: Warning: Couldn't get"
+                                 "%s\n", url);
+                    }
+               }
+          }
      }
 
      D("New finished\n");
@@ -1065,7 +1228,7 @@ NPError NPP_Destroy(NPP instance, NPSavedData** save)
 {
      int e;
 
-     D("Destroy\n");
+     D("Destroy() - instance=%p\n", instance);
 
      if (!instance)
 	  return NPERR_INVALID_INSTANCE_ERROR;
@@ -1088,9 +1251,8 @@ NPError NPP_Destroy(NPP instance, NPSavedData** save)
 	  NPN_MemFree((char *)THIS->args);
 
 	  free(THIS->mimetype);
-	  free(THIS->href);
-	  free(THIS->mms);
-
+	  THIS->href = NULL;
+	  THIS->mms = NULL;
 	  NPN_MemFree(THIS);
 	  THIS = NULL;
      }
@@ -1115,25 +1277,28 @@ static void new_child(NPP instance, const char* fname)
      if (THIS->pid != -1)
 	  return;
 
+     /* Guard against spawning helper if no command! */
+     if(THIS->command == 0)
+         return;
+
      if (socketpair(AF_UNIX, SOCK_STREAM, 0, pipe) < 0)
      {
 	  NPN_Status(instance, "MozPlugger: Failed to create a pipe!");
 	  return;
      }
 
-     if ((THIS->pid = my_fork(instance)) == -1)
+     if ((THIS->pid = my_fork(instance, pipe[1])) == -1)
+     {
+	  NPN_Status(instance, "MozPlugger: Failed to fork helper!");
 	  return;
-    
+     }
+     
      if (!THIS->pid)
      {
-	  if (!find_command(instance, 1) && !find_command(instance, 0))
-	       _exit(EX_UNAVAILABLE);
-
-	  THIS->fd = pipe[1];
 	  close(pipe[0]);
 
 	  D("CHILD RUNNING run() [2]\n");
-	  run(instance, fname);
+	  run(instance, fname, pipe[1]);
 
 	  _exit(EX_UNAVAILABLE);
      }
@@ -1154,13 +1319,18 @@ NPError NPP_NewStream(NPP instance,
 		      NPBool seekable,
 		      uint16 *stype)
 {
-     D("NewStream\n");
+     char * savedMimetype = NULL;
+
+     D("NewStream() - instance=%p\n", instance);
 
      if (instance == NULL)
 	  return NPERR_INVALID_INSTANCE_ERROR;
 
      if (THIS->pid != -1)
+     {
+          D("NewStream() exiting process already running\n");
 	  return NPERR_INVALID_INSTANCE_ERROR;
+     }
 
      /* This is a stupid special case and should be coded into
       * mozpluggerc instead... */
@@ -1181,14 +1351,41 @@ NPError NPP_NewStream(NPP instance,
      D("Mime type %s\n", type);
      D("Url is %s (seekable=%d)\n", stream->url, seekable);
 
-     /* A command may be removed after mozplugger has been loaded. */
-     if (!find_command(instance, 1) && !find_command(instance, 0))
+     /* Ocassionally the MIME type here is different to that passed to the
+      * NEW function - this is because of either badly configure web server
+      * who's HTTP response content-type does not match the mimetype in the
+      * preceding embebbed, object or link tag. Or badly constructed embedded
+      * tag or ambiguity in the file extension to mime type mapping. Lets
+      * first assume the HTTP response was correct and if not fall back to
+      * the original tag in the mime type. */
+     if(strcmp(type, THIS->mimetype) != 0)
      {
-	  NPN_Status(instance, "MozPlugger: No appropriate application found.");
-	  return NPERR_GENERIC_ERROR;
+         D("Mismatching mimetype reported, originally was \'%s\' now '\%s' "
+           "for url %s\n", THIS->mimetype, type, stream->url);
+         savedMimetype = THIS->mimetype;
+         THIS->mimetype = strdup(type);
      }
 
-     if ((THIS->flags & H_STREAM)
+     /* Find a suitable command, fall back to savedMimeType if defined */
+     while(!find_command(instance, 1) && !find_command(instance, 0))
+     {
+          if(savedMimetype)       /* Lets try the saved one */
+          {
+              free(THIS->mimetype);
+              THIS->mimetype = savedMimetype;
+              savedMimetype = NULL;
+          }
+          else 
+          {
+	      NPN_Status(instance, 
+                         "MozPlugger: No appropriate application found.");
+	      return NPERR_GENERIC_ERROR;
+          }
+     }
+     free(savedMimetype);     /* discard the saved Mime type */
+
+
+     if ((THIS->cmd_flags & H_STREAM)
 	 && strncasecmp(stream->url, "file:", 5)
 	 && strncasecmp(stream->url, "imap:", 5)
 	 && strncasecmp(stream->url, "mailbox:", 8))
@@ -1211,36 +1408,59 @@ void NPP_StreamAsFile(NPP instance,
 		      NPStream *stream,
 		      const char* fname)
 {
-     D("StreamAsFile\n");
+     D("StreamAsFile() - instance=%p\n", instance);
      new_child(instance, fname);
 }
 
+/*****************************************************************************
+ * The browser should have resized the window for us, but this function was
+ * added because of a bug in Mozilla 1.7 (see Mozdev bug #7734) and
+ * https://bugzilla.mozilla.org/show_bug.cgi?id=201158
+ *
+ * Bug was fixed in Mozilla CVS repositary in version 1.115 of
+ * ns4xPluginInstance.cpp (13 Nov 2003), at the time version was 0.13.
+ * version 0.14 happened on 14th July 2004
+ ****************************************************************************/
 void resize_window(NPP instance)
 {
+     /* Mozilla work around not require for versions > 0.13 */
+     if((browserApiMajorVer > 0) || (browserApiMinorVer > 13))
+         return;
+
      XSetWindowAttributes attrib;
      attrib.override_redirect = True;
      XChangeWindowAttributes(THIS->display, (Window)THIS->windata.window,
-			     CWOverrideRedirect, &attrib);
+			     (unsigned long) CWOverrideRedirect, &attrib);
      
-     D("Resizing WIN 0x%x to %dx%d!?\n", THIS->windata.window,
-       THIS->windata.width, THIS->windata.height);
+     D("Bug #7734 work around - resizing WIN 0x%x to %dx%d!?\n", 
+            THIS->windata.window,
+            THIS->windata.width, THIS->windata.height);
+
      XResizeWindow(THIS->display, (Window)THIS->windata.window,
-		   THIS->windata.width, THIS->windata.height);
+		   (unsigned) THIS->windata.width, 
+                   (unsigned) THIS->windata.height);
 }
 
 /*****************************************************************************
- * Called when initializing or resizing the window.
- *****************************************************************************/
+ * The browser calls NPP_SetWindow after creating the instance to allow drawing
+ * to begin. Subsequent calls to NPP_SetWindow indicate changes in size or 
+ * position. If the window handle is set to null, the window is destroyed. In
+ * this case, the plug-in must not perform any additional graphics operations
+ * on the window and should free any associated resources.
+  ***************************************************************************/
 NPError NPP_SetWindow(NPP instance, NPWindow* window)
 {
-     D("SetWindow\n");
+     D("SetWindow() - instance=%p\n", instance);
 
      if (!instance)
 	  return NPERR_INVALID_INSTANCE_ERROR;
- 
+
      if (!window)
 	  return NPERR_NO_ERROR;
  
+     /* TODO - Should really pass to helper / controller the fact that the
+      * window has disappeared! - need to check the consequences first
+      * before deleting these 2 lines */
      if (!window->window)
 	  return NPERR_NO_ERROR;
 
@@ -1250,11 +1470,22 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
      THIS->display = ((NPSetWindowCallbackStruct *)window->ws_info)->display;
      THIS->displayname = XDisplayName(DisplayString(THIS->display));
      THIS->windata = *window;
-     
+
      if (THIS->mms)
      {
+          if(THIS->command == 0)
+          {
+               /* Can only use streaming commands, as Mozilla cannot handle
+                * these types (mms) of urls */
+               if (!find_command(instance, 1))
+               {
+	           THIS->mms = NULL;
+	           NPN_Status(instance, 
+                          "MozPlugger: No appropriate application found.");
+ 	           return NPERR_GENERIC_ERROR;
+               }
+          }
 	  new_child(instance, THIS->mms);
-	  free(THIS->mms);
 	  THIS->mms = NULL;
 	  return NPERR_NO_ERROR;
      }
@@ -1269,7 +1500,6 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
 
      /* In case Mozilla would call NPP_SetWindow() in a loop. */
      usleep(4000);
-
      return NPERR_NO_ERROR;
 }
 
@@ -1278,13 +1508,27 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
  *****************************************************************************/
 NPError NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
 {
-     D("DestroyStream\n");
+     D("DestroyStream() - instance=%p\n", instance);
      return NPERR_NO_ERROR;
 }
 
+/*****************************************************************************
+ * The browser calls this function only once; when the plug-in is loaded, 
+ * before the first instance is created. NPP_Initialize tells the plug-in that
+ * the browser has loaded it.
+ *****************************************************************************/
 NPError NPP_Initialize(void)
 {
-     D("Initialize\n");
+     int pluginApiMajorVer;
+     int pluginApiMinorVer;
+
+     NPN_Version(&pluginApiMajorVer,  &pluginApiMinorVer,
+                 &browserApiMajorVer, &browserApiMinorVer);
+
+     D("Initialize - API versions plugin=%d.%d Browser=%d.%d\n", 
+             pluginApiMajorVer, pluginApiMinorVer,
+             browserApiMajorVer, browserApiMinorVer);
+
      do_read_config();
      return NPERR_NO_ERROR;
 }
@@ -1302,7 +1546,7 @@ void NPP_Shutdown(void)
 
 void NPP_Print(NPP instance, NPPrint* printInfo)
 {
-     D("Print\n");
+     D("Print() - instance=%p\n", instance);
 }
 
 int32 NPP_Write(NPP instance,
@@ -1311,23 +1555,19 @@ int32 NPP_Write(NPP instance,
 		int32 len,
 		void *buf)
 {
-     D("Write %d\n", len);
+     D("Write(%d) - instance=%p\n", len, instance);
      return len;
 }
 
 int32 NPP_WriteReady(NPP instance, NPStream *stream)
 {
+    D("WriteReady() - instance=%p\n", instance);
     if (instance)
     {
 	 /* Tell the browser that it can finish with the stream
 	    (actually we just wanted the name of the stream!)
 	    And not the stream data. */
         NPN_DestroyStream(instance, stream, NPRES_DONE);
-        D("NPP_WriteReady - terminating the stream\n");
-    }
-    else 
-    {
-        D("NPP_WriteReady called with instance=0\n");
     }
     return 0;
 }
