@@ -70,6 +70,7 @@
 #define H_FILL 		0x400
 #define H_HIDDEN        0x800
 #define H_CONTROLS 	0x1000
+#define H_LINKS         0x2000
 
 /*****************************************************************************
  * mozplugger-common.c functions
@@ -78,4 +79,4 @@
 void close_debug();
 #endif
 void D(char *fmt, ...);
-void my_kill(int pid);
+void my_kill(pid_t pid);
